/*****************************************************************************
Original version of mpeg2encode is copyrighted to:
MPEG Software Simulation Group. All Rights Reserved.

Disclaimer of Warranty

mpeg2encode is available to the user without any license fee or
royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
any and all warranties, whether express, implied, or statuary, including any
implied warranties or merchantability or of fitness for a particular
purpose.  In no event shall the copyright-holder be liable for any
incidental, punitive, or consequential damages of any kind whatsoever
arising from the use of these programs.

This disclaimer of warranty extends to the user of these programs and user's
customers, employees, agents, transferees, successors, and assigns.

The MPEG Software Simulation Group does not represent or warrant that the
programs furnished hereunder are free of infringement of any third-party
patents.

Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
are subject to royalty fees to patent holders.  Many of these patents are
general enough such that they are unavoidable regardless of implementation
design.


Modification of mpeg2encode contained in PM2V source code is copyrighted by 
Jean Laurent and Olivier Guillot.
This modification are distributed under the GPL (GNU Public License). 
see license.txt in the doc directory for details.

You can contact us to pm2v@free.fr or pm2v@sourceforge.net
*****************************************************************************/


#ifndef PM2V
#include <stdio.h>
#include <mpi.h>


typedef MPI_Comm pm2v_communicator;

#define HOR_SIZE_RANGE 1
#define HOR_SIZE_MOD_4096 2
#define HOR_SIZE_NOT_MOD_2 3
#define VERT_SIZE_RANGE 4
#define VERT_SIZE_MOD_4096 5
#define VERT_SIZE_NOT_MOD_2 6
#define VERT_SIZE_NOT_MOD_4 7
#define PEL_ASPECT_RANGE 8
#define BAD_ASPECT_RATIO 9
#define BAD_FRAME_RATE 10
#define BITRATE_RANGE 11
#define VBV_BUFFER_SIZE_RANGE 12
#define BAD_CHROMA_FORMAT 13
#define BAD_VIDEO_FORMAT 14
#define BAD_COLOR_PRIMARIES 15
#define TRANSFER_CHARA_RANGE 16
#define MATRIX_COEFF 17
#define OUTPUT_HOR_SIZE_RANGE 18
#define OUTPUT_VERT_SIZE_RANGE 19
#define INTRA_DC_PREC_RANGE 20
#define F_CODE_RANGE 21
#define NEGATIVE_SEARCH_WIN 22
#define BAD_PROFILE_RANGE 23
#define BAD_PROFILE 24
#define BAD_LEVEL 25
#define BAD_PROFILE_LEVEL 26
#define B_PICTURE_FORBIDDEN 27
#define BAD_CHROMA_FORMAT_PROFILE 28
#define NOT_REPEAT_FIRST_ZERO 29
#define BAD_DC_PRECISION 30
#define UNSUPPORTED 31
#define BAD_N 32

static const struct {
  int code;
  char * name;
}pm2v_errorarray[]=
{{HOR_SIZE_RANGE,"Horizontal size exceeds allowed range"},
 {HOR_SIZE_MOD_4096,"Horizontal size must not be a multiple of 4096"},
 {HOR_SIZE_NOT_MOD_2,"Horizontal size must be an even"},
 {VERT_SIZE_RANGE,"Vertical size exceeds allowed range"},
 {VERT_SIZE_MOD_4096,"Horizontal size must not be a multiple of 4096"},
 {VERT_SIZE_NOT_MOD_2,"Vertical size must be an even"},
 {VERT_SIZE_NOT_MOD_4,"Vertical size must be a multiple of 4"},
 {PEL_ASPECT_RANGE,"Pel aspect ratio exceeds allowed range"},
 {BAD_ASPECT_RATIO,"Unknown frame rate code"},
 {BITRATE_RANGE,"Bit rate exceeds allowed range"},
 {VBV_BUFFER_SIZE_RANGE,"VBV buffer size exceeds allowed range"},
 {BAD_CHROMA_FORMAT,"Unknown chroma format"},
 {BAD_VIDEO_FORMAT,"Unknown video format"},
 {BAD_COLOR_PRIMARIES,"Color primaries must be in range 1...2 or 4...7"},
 {TRANSFER_CHARA_RANGE,"Transfer characteristics must be in range 1...2 or 4...7"},
 {MATRIX_COEFF,"Matrix coefficients must be in range 1...2 or 4...7"},
 {OUTPUT_HOR_SIZE_RANGE,"Display horizontal size must be in range 0...16383"},
 {OUTPUT_VERT_SIZE_RANGE,"Display vertical_size must be in range 0...16383"},
 {INTRA_DC_PREC_RANGE,"Intra dc precision must be in range 0...3"},
 {F_CODE_RANGE,"F code exceeds allowed range"},
 {NEGATIVE_SEARCH_WIN,"Search window must be positive"},
 {BAD_PROFILE,"Unknown profile"},
 {BAD_LEVEL,"Unknown level"},
 {UNSUPPORTED,"Unsupported feature"},
 {BAD_PROFILE_LEVEL,"Undefined profile@level combination"},
 {B_PICTURE_FORBIDDEN,"Simple Profile does not allow B pictures"},
 {BAD_CHROMA_FORMAT_PROFILE,"Bad chroma format for specified profile"},
 {NOT_REPEAT_FIRST_ZERO,"Repeat_first_first must be zero"},
 {BAD_DC_PRECISION,"11 bit DC precision only allowed in High Profile"},
 {BAD_N,"Invalid N"}};
  
 
 
 


typedef struct pm2v_param_ {
  
  FILE* output;
  /* reconstructed frames */
  unsigned char *newrefframe[3], *oldrefframe[3], *auxframe[3];
  /* original frames */
  unsigned char *neworgframe[3], *oldorgframe[3], *auxorgframe[3];
  /* prediction of current frame */
  unsigned char *predframe[3];
  /* 8*8 block data */
  short (*blocks)[64];
  /* intra / non_intra quantization matrices */
  unsigned char intra_q[64], inter_q[64];
  unsigned char chrom_intra_q[64],chrom_inter_q[64];
  /* prediction values for DCT coefficient (0,0) */
  int dc_dct_pred[3];
  /* macroblock side information array */
  struct mbinfo *mbinfo;
  /* motion estimation parameters */
  struct motion_data *motion_data;
  /* clipping (=saturation) table */
  unsigned char *clp;
  
  /* name strings */
  char tplorg[256], tplref[256];
  char iqname[256], niqname[256];
  char * id_string;
  // char statname[256];
  // char errortext[256];

  //FILE *output, *statfile; /* file descriptors */
  int inputtype; /* format of input frames */

  int quiet; /* suppress warnings */


  /* coding model parameters */

  int N; /* number of frames in Group of Pictures */
  int M; /* distance between I/P frames */
  int P; /* intra slice refresh interval */
  int nframes; /* total number of frames to encode */
  int frame0, tc0; /* number and timecode of first frame */
  int mpeg1; /* ISO/IEC IS 11172-2 sequence */
  int fieldpic; /* use field pictures */

  /* sequence specific data (sequence header) */

  int horizontal_size, vertical_size; /* frame size (pels) */
  int width, height; /* encoded frame size (pels) multiples of 16 or 32 */
  int chrom_width,chrom_height,block_count;
  int mb_width, mb_height; /* frame size (macroblocks) */
  int width2, height2, mb_height2, chrom_width2; /* picture size */
  int aspectratio; /* aspect ratio information (pel or display) */
  int frame_rate_code; /* coded value of frame rate */
  double frame_rate; /* frames per second */
  double bit_rate; /* bits per second */
  int vbv_buffer_size; /* size of VBV buffer (* 16 kbit) */
  int constrparms; /* constrained parameters flag (MPEG-1 only) */
  int load_iquant, load_niquant; /* use non-default quant. matrices */
  int load_ciquant,load_cniquant;


/* sequence specific data (sequence extension) */

  int profile, level; /* syntax / parameter constraints */
  int prog_seq; /* progressive sequence */
  int chroma_format;
  int low_delay; /* no B pictures, skipped pictures */


/* sequence specific data (sequence display extension) */

  int video_format; /* component, PAL, NTSC, SECAM or MAC */
  int color_primaries; /* source primary chromaticity coordinates */
  int transfer_characteristics; /* opto-electronic transfer char. (gamma) */
  int matrix_coefficients; /* Eg,Eb,Er / Y,Cb,Cr matrix coefficients */
  int display_horizontal_size, display_vertical_size; /* display size */


/* picture specific data (picture header) */

  int temp_ref; /* temporal reference */
  int pict_type; /* picture coding type (I, P or B) */
  int vbv_delay; /* video buffering verifier delay (1/90000 seconds) */

  int r,Xi,Xb,Xp,d0i,d0p,d0b; /* rate control */
  double avg_act; /* rate control */

/* picture specific data (picture coding extension) */

  int forw_hor_f_code, forw_vert_f_code;
  int back_hor_f_code, back_vert_f_code; /* motion vector ranges */
  int dc_prec; /* DC coefficient precision for intra coded blocks */
  int pict_struct; /* picture structure (frame, top / bottom field) */
  int topfirst; /* display top field first */
/* use only frame prediction and frame DCT (I,P,B,current) */
  int frame_pred_dct_tab[3], frame_pred_dct;
  int conceal_tab[3]; /* use concealment motion vectors (I,P,B) */
  int qscale_tab[3], q_scale_type; /* linear/non-linear quantizaton table */
  int intravlc_tab[3], intravlc; /* intra vlc format (I,P,B,current) */
  int altscan_tab[3], altscan; /* alternate scan (I,P,B,current) */
  int repeatfirst; /* repeat first field after second field */
  int prog_frame; /* progressive frame */
} pm2v_param;


typedef struct pm2v_stream_ {
  unsigned char *** pictures_pool;
  pm2v_communicator com;
  int local_error;
  int master_id;
  int my_id;
  int proc_number;
  // int * nbMB;  /* Array of number of macroblock for each process */
  // int * stride; /* Well ... stride */
  // int * sigrid; 
  // int * stride_blocks;
  int * macroblock_to_process;
  int * offset_macroblock;
  //int * offset_block;
  int nb_vmacroblock;
  //int * nb_blocks;
  //int * nb_chroma[3];
  int * nb_macroblock;

  int * nb_lines;
  //int * nb_linesc;
  int * offset_lines;
  //int * offset_linesc;
  //int * nb_macroblock2;
  //int * nb_lumi[3];
  //int * stride_chroma[3];
  //int * stride_lumi[3];
  int * idx_macroblock;
  char * buffcopy;
  char * buffcopy2;
  unsigned char * neworg;
  double * rate_mcomp;
  double * time_putpic;
  int picture_count;
  MPI_Datatype type_chroma_block420;
  MPI_Datatype type_chroma_block422;
  MPI_Datatype type_lumi_block;
  MPI_Datatype type_macroblock;
  MPI_Datatype type_macroblock2;
  MPI_Datatype type_block;
  pm2v_param param;
} pm2v_stream;

void pm2v_broadcastparam(pm2v_stream* stream,char * param);
void pm2v_getparam(pm2v_stream* stream,char ** param);
void pm2v_barrier(pm2v_stream * stream);
void pm2v_broadcast_image(pm2v_stream* stream, char * src);
void pm2v_reduce_macroblock(pm2v_stream* stream);
void pm2v_reduce_lumi(pm2v_stream * stream,char * src);
void pm2v_reduce_chroma(pm2v_stream * stream,char * src,int chroma);
void pm2v_init_com(pm2v_stream* stream);
void pm2v_bcast_load(pm2v_stream * stream);
void pm2v_gather_time(pm2v_stream * stream);
void pm2v_scatter_image(pm2v_stream * stream,char * src,int width,int height);
void pm2v_gather_image(pm2v_stream * stream,char * src,int width,int height);
/* Interface */
int pm2v_open(pm2v_stream* stream, char * param,FILE * output,pm2v_communicator com,int master_id,int my_id,int nb_proc);
void pm2v_add_image(pm2v_stream* stream,unsigned char* img);
void pm2v_close(pm2v_stream* stream);



/* a virer */

void pm2v_readppm(pm2v_stream* stream,char* filename, unsigned char* img);

#define PM2V
#endif







