/*****************************************************************************
Original version of mpeg2encode is copyrighted to:
MPEG Software Simulation Group. All Rights Reserved.

Disclaimer of Warranty

mpeg2encode is available to the user without any license fee or
royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
any and all warranties, whether express, implied, or statuary, including any
implied warranties or merchantability or of fitness for a particular
purpose.  In no event shall the copyright-holder be liable for any
incidental, punitive, or consequential damages of any kind whatsoever
arising from the use of these programs.

This disclaimer of warranty extends to the user of these programs and user's
customers, employees, agents, transferees, successors, and assigns.

The MPEG Software Simulation Group does not represent or warrant that the
programs furnished hereunder are free of infringement of any third-party
patents.

Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
are subject to royalty fees to patent holders.  Many of these patents are
general enough such that they are unavoidable regardless of implementation
design.


Modification of mpeg2encode contained in PM2V source code is copyrighted by 
Jean Laurent and Olivier Guillot.
This modification are distributed under the GPL (GNU Public License). 
see license.txt in the doc directory for details.

You can contact us to pm2v@free.fr or pm2v@sourceforge.net
*****************************************************************************/

#include  "fdctref.h"
#include  "idct.h"
#include  "motion.h"
#include  "mpeg2enc.h"
#include  "params.h"
#include  "predict.h"
#include  "putbits.h"
#include  "puthdr.h"
#include  "putmpg.h"
#include  "putpic.h"
#include  "putseq.h"
#include  "putvlc.h"
#include  "quantize.h"
#include  "ratectl.h"
#include  "readpic.h"
#include  "stats.h"
#include  "transfrm.h"
#include  "writepic.h"
#include <stdio.h>

#include <assert.h>
#include "config.h"
#include "quant.h"
#include "vlc.h"

/* private prototypes */
static void putDC (pm2v_stream * stream, sVLCtable * tab, int val);

/* generate variable length code for luminance DC coefficient */
void 
putDClum (pm2v_stream * stream, int val)
{
  putDC (stream, DClumtab, val);
}

/* generate variable length code for chrominance DC coefficient */
void 
putDCchrom (pm2v_stream * stream, int val)
{
  putDC (stream, DCchromtab, val);
}

/* generate variable length code for DC coefficient (7.2.1) */
static void 
putDC (pm2v_stream * stream, sVLCtable * tab, int val)
{
  int absval, size;

  absval = (val < 0) ? -val : val;	/* abs(val) */

  assert (!(absval > 2047 || (stream->param.mpeg1 && absval > 255)));

  //sprintf(errortext,"DC value out of range (%d)\n",val);

  /* compute dct_dc_size */
  size = 0;

  while (absval)
    {
      absval >>= 1;
      size++;
    }

  /* generate VLC for dct_dc_size (Table B-12 or B-13) */
  putbits (stream, tab[size].code, tab[size].len);

  /* append fixed length code (dc_dct_differential) */
  if (size != 0)
    {
      if (val >= 0)
	absval = val;
      else
	absval = val + (1 << size) - 1;		/* val + (2 ^ size) - 1 */
      putbits (stream, absval, size);
    }
}

/* generate variable length code for first coefficient
 * of a non-intra block (7.2.2.2) */
void 
putACfirst (pm2v_stream * stream, int run, int val)
{
  if (run == 0 && (val == 1 || val == -1))	/* these are treated differently */
    putbits (stream, 2 | (val < 0), 2);		/* generate '1s' (s=sign), (Table B-14, line 2) */
  else
    putAC (stream, run, val, 0);	/* no difference for all others */
}

/* generate variable length code for other DCT coefficients (7.2.2) */
void 
putAC (pm2v_stream * stream, int run, int signed_level, int vlcformat)
{
  int level, len;
  VLCtable *ptab = NULL;

  level = (signed_level < 0) ? -signed_level : signed_level;	/* abs(signed_level) */

  /* make sure run and level are valid */
  assert (!(run < 0 || run > 63 || level == 0 || level > 2047 || (stream->param.mpeg1 && level > 255)));

  // sprintf(errortext,"AC value out of range (run=%d, signed_level=%d)\n",run,signed_level);


  len = 0;

  if (run < 2 && level < 41)
    {
      /* vlcformat selects either of Table B-14 / B-15 */
      if (vlcformat)
	ptab = &dct_code_tab1a[run][level - 1];
      else
	ptab = &dct_code_tab1[run][level - 1];

      len = ptab->len;
    }
  else if (run < 32 && level < 6)
    {
      /* vlcformat selects either of Table B-14 / B-15 */
      if (vlcformat)
	ptab = &dct_code_tab2a[run - 2][level - 1];
      else
	ptab = &dct_code_tab2[run - 2][level - 1];

      len = ptab->len;
    }

  if (len != 0)			/* a VLC code exists */
    {
      putbits (stream, ptab->code, len);
      putbits (stream, signed_level < 0, 1);	/* sign */
    }
  else
    {
      /* no VLC for this (run, level) combination: use escape coding (7.2.2.3) */
      putbits (stream, 1l, 6);	/* Escape */
      putbits (stream, run, 6);	/* 6 bit code for run */
      if (stream->param.mpeg1)
	{
	  /* ISO/IEC 11172-2 uses a 8 or 16 bit code */
	  if (signed_level > 127)
	    putbits (stream, 0, 8);
	  if (signed_level < -127)
	    putbits (stream, 128, 8);
	  putbits (stream, signed_level, 8);
	}
      else
	{
	  /* ISO/IEC 13818-2 uses a 12 bit code, Table B-16 */
	  putbits (stream, signed_level, 12);
	}
    }
}

/* generate variable length code for macroblock_address_increment (6.3.16) */
void 
putaddrinc (pm2v_stream * stream, int addrinc)
{
  while (addrinc > 33)
    {
      putbits (stream, 0x08, 11);	/* macroblock_escape */
      addrinc -= 33;
    }

  putbits (stream, addrinctab[addrinc - 1].code, addrinctab[addrinc - 1].len);
}

/* generate variable length code for macroblock_type (6.3.16.1) */
void 
putmbtype (pm2v_stream * stream, int pict_type, int mb_type)
{
  putbits (stream, mbtypetab[pict_type - 1][mb_type].code,
	   mbtypetab[pict_type - 1][mb_type].len);
}

/* generate variable length code for motion_code (6.3.16.3) */
void 
putmotioncode (pm2v_stream * stream, int motion_code)
{
  int abscode;

  abscode = (motion_code >= 0) ? motion_code : -motion_code;	/* abs(motion_code) */
  putbits (stream, motionvectab[abscode].code, motionvectab[abscode].len);
  if (motion_code != 0)
    putbits (stream, motion_code < 0, 1);	/* sign, 0=positive, 1=negative */
}

/* generate variable length code for dmvector[t] (6.3.16.3), Table B-11 */
void 
putdmv (pm2v_stream * stream, int dmv)
{
  if (dmv == 0)
    putbits (stream, 0, 1);
  else if (dmv > 0)
    putbits (stream, 2, 2);
  else
    putbits (stream, 3, 2);
}

/* generate variable length code for coded_block_pattern (6.3.16.4)

 * 4:2:2, 4:4:4 not implemented
 */
void 
putcbp (pm2v_stream * stream, int cbp)
{
  putbits (stream, cbptable[cbp].code, cbptable[cbp].len);
}
